#include <mask.h>
#include <printapp.h>
#include <tabutil.h>
#include <utility.h>
#include <lffiles.h>

#include "at7.h"

// nomi dei campi
#include "soggetti.h"

#define ALIAS_LCPD	100
#define ALIAS_LCPR	200
#define ALIAS_COMD	300
#define ALIAS_COMR	400

class TAggiornaCAP : public TPrintapp
{
	TMask*						_msk;
	TRelation*   			_rel;
	int 							_cur;	
	bool							_scrivi;

  static bool filter_func_cap(const TRelation* rel);

protected:
	virtual bool user_create();
	virtual bool user_destroy();
	virtual bool set_print(int m);
	virtual void set_page(int file, int cnt);
	virtual bool preprocess_page(int file, int counter);
	
public:
	TAggiornaCAP() {}
};

HIDDEN inline TAggiornaCAP& app() { return (TAggiornaCAP&) main_app(); }

bool TAggiornaCAP::filter_func_cap(const TRelation* rel)
{   
	bool filtrato = FALSE;
	TString16 cap;
	TString16 loc;
	TString16 capcom;
	TRectype recsog = rel->lfile().curr();
	cap = recsog.get(SOG_DOM_CAP);
	loc = recsog.get(SOG_DOM_CODLOC);
	if (loc.empty())
		capcom = rel->lfile(-ALIAS_COMD).get("CAPCOM");
	else
		capcom = rel->lfile(-ALIAS_LCPD).get("S6");	
	if (capcom.not_empty() && cap != capcom)
		filtrato = TRUE;
	
	cap = recsog.get(SOG_RES_CAP);
	loc = recsog.get(SOG_RES_CODLOC);
	if (loc.empty())
		capcom = rel->lfile(-ALIAS_COMR).get("CAPCOM");
	else
		capcom = rel->lfile(-ALIAS_LCPR).get("S6");	
	if (capcom.not_empty() && cap != capcom)
		filtrato = TRUE;
	return filtrato;	
}

bool TAggiornaCAP::preprocess_page(int file, int counter)
{ 
	bool rew = FALSE;
	TString16 cap;
	TString16 loc;
	TString16 capcom;
	TRectype& recsog = current_cursor()->curr();
	cap = recsog.get(SOG_DOM_CAP);
	loc = recsog.get(SOG_DOM_CODLOC);
	if (loc.empty())
		capcom = current_cursor()->file(-ALIAS_COMD).get("CAPCOM");
	else
		capcom = current_cursor()->file(-ALIAS_LCPD).get("S6");	
	if (capcom.not_empty() && (cap != capcom))
		recsog.put(SOG_DOM_CAP, capcom);
	cap = recsog.get(SOG_RES_CAP);
	loc = recsog.get(SOG_RES_CODLOC);
	if (loc.empty())
		capcom = current_cursor()->file(-ALIAS_COMR).get("CAPCOM");
	else
		capcom = current_cursor()->file(-ALIAS_LCPR).get("S6");	
	if (capcom.not_empty() && (cap != capcom))
		recsog.put(SOG_RES_CAP, capcom);
	if (_scrivi)
		current_cursor()->file().rewrite();
  return TRUE;
}

void TAggiornaCAP::set_page(int file, int cnt)
{ 
	set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
	set_row(1,"@10g@S", FLD(LF_SOGGETTI,SOG_COGNOME));
	set_row(1,"@36g@S", FLD(LF_SOGGETTI,SOG_NOME));
	set_row(1,"@62g@S", FLD(LF_SOGGETTI,SOG_CODSEZ));
	set_row(1,"@67g@S", FLD(LF_SOGGETTI,SOG_DOM_CAP));
	set_row(1,"@74g@S", FLD(LF_SOGGETTI,SOG_RES_CAP));
	set_row(1,"@81g@ld", FLD(LF_SOGGETTI,SOG_DATAULTAGG));
	set_row(1,"@92g@S", FLD(LF_SOGGETTI,SOG_UTENULTAGG));
}

bool TAggiornaCAP::set_print(int)
{
  KEY tasto;
  tasto = _msk->run();
	if (tasto == K_ENTER)
	{  	                      
		_scrivi = !_msk->get_bool(102);
    reset_files(); 
    add_file(LF_SOGGETTI);
		if (_msk->get_bool(101))
			current_cursor()->setfilter("(90->DOM_CODLOC != \"\")", TRUE);
		current_cursor()->set_filterfunction (filter_func_cap, FALSE);
		reset_print();		
    return TRUE;
  }
  else
    return FALSE;
}

bool TAggiornaCAP::user_create()
{
	_msk = new TMask("at7100a");
	_rel = new TRelation(LF_SOGGETTI);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCPD);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,0,ALIAS_COMD);
  _rel->add("LCP", "CODTAB==RES_CODLOC",1,0,ALIAS_LCPR);
  _rel->add(LF_COMUNI, "COM==RES_CODCOM",1,0,ALIAS_COMR);
  _cur = add_cursor(new TCursor(_rel, "", 1));
	return TRUE;
}	

bool TAggiornaCAP::user_destroy()	
{
	delete _rel;
	delete _msk;
	return TRUE;
}

int at7100(int argc, char* argv[])
{
	TAggiornaCAP a;
	a.run(argc, argv, "Aggiornamento CAP su indirizzi");
	return 0;
}